<?php
// $Id: discogs.import.inc,v 1.1 2011/01/19 16:15:56 karlheinz Exp $
/**
 * @file
 * Functions for importing releases from Discogs.com.
 */

/**
 * Form for importing releases from Discogs.com into Drupal. This is the first
 * form stage, where the user enters a keyword and type to fetch a list of
 * releases (or, if that fails, a list of artists/labels to resubmit).
 * Those releases will be displayed in subsequent form stages, generated by
 * discogs_import_releases_form().
 *
 * @param $form_state
 *      Form state
 * @return
 *      Multidimensional $form array
 * @ingroup forms
 * @see discogs_import_releases_form()
 */
function discogs_import_form(&$form_state) {
  // Display results
  if (isset($form_state['storage']['step'])
  && ($form_state['storage']['step'] > 0)
  && is_array($form_state['storage']['data'])) {
    return discogs_import_releases_form($form_state);
  }
  $form_state['storage']['step'] = 0;

  $form['term'] = array(
    '#type' => 'textfield',
    '#title' => t('Release, Artist Name, or Label Name'),
    '#description' => t('Release, Artist, or Label to pull from Discogs.com.
      If you know the exact release ID, artist name, or label name, you will
      be presented with all the relevant releases to import. If not, a search
      will be performed, and you may then choose which releases you want to
      import. If your search turns up unexpected results, try searching with
      a more specific phrase.'),
  );
  $form['type'] = array(
    '#type' => 'select',
    '#default_value' => 'release',
    '#options' => array(
      'release' => t('Release'),
      'artist' => t('Artist'),
      'label' => t('Label'),
  ),
  );
  // Custom submit
  $form['next'] = array(
        '#type' => 'submit',
        '#value' => 'Next >>',
  );

  return $form;
}

/**
 * This function generates the subsequent form stages for importing releases
 * from Discogs.com. At this point, the user has either returned an array of
 * releases, or an array of data to resubmit. In either case, it will be
 * presented to the user as a checkbox. The form submit function will work
 * out whether to submit it, or re-display it with fetched releases.
 *
 * @param $form_state
 *      Form state
 * @return
 *      Multidimensional $form array
 * @ingroup forms
 * @see discogs_import_form()
 */
function discogs_import_releases_form($form_state) {
  // Informational form element
  if (isset($form_state['storage']['info']) && !empty($form_state['storage']['info'])) {
    $form['info'] = array(
      '#value' => theme_discogs_forminfo($form_state['storage']['info']),
    );
  }
  // Format releases array for checkbox
  $chex = array();
  foreach ($form_state['storage']['data'] as $did => $data) {
    $text = theme('discogs_checkbox_text', $data);
    $id = isset($data['id']) ? $data['id'] : $did;
    $chex[$id] = $text;
  }
  // Build check boxes
  $type = ($form_state['storage']['step'] == 2) ? 'release' : $form_state['storage']['type'];
  $form['data'] = array(
    '#type' => 'checkboxes',
    '#title' => t('Returned releases'),
    '#options' => $chex,
  );
  // Set the #value of our submit depending upon the import stage
  $value = ($form_state['storage']['step'] == 2) ? 'Import' : 'Get';
  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => $value,
  );
  return $form;
}

/**
 * Submit handler for discogs_import_form().
 *
 * Will decide at what step of the form process we are at, and set form
 * variables accordingly:
 * 0 = First submitted
 * 1 = Searched, returned releases
 * 2 = Final import of releases
 *
 * @param $form
 *   Form array
 * @param $form_state
 *   Form state
 * @return
 */
function discogs_import_form_submit($form, &$form_state) {
  $api_key  = variable_get('discogs_api_key', 'ae26686b33');

  // Handle Step 0 (First submitted)
  if (!isset($form_state['storage']['step']) || ($form_state['storage']['step'] < 1)) {
    // Get data from Discogs.com
    $term     = urlencode($form_state['values']['term']);
    $type     = $form_state['values']['type'];
    $releases = discogs_fetch_release($term, $type, $api_key);
    if ($releases) {
      // Something was fetched

      // $releases array will be different if it returned a single release
      if ($type == 'release') {
        $form_state['storage']['data'] = array($releases['release']);
      }
      else {
        $form_state['storage']['data'] = $releases[$type]['releases']['release'];
      }
      $form_state['storage']['step'] = 2;
      $form_state['storage']['type'] = $type;
      // Informational elements
      $form_state['storage']['info'] = $releases[$type];
      unset($form_state['storage']['info']['releases']);
    }
    if (is_null($form_state['storage']['data']) || empty($form_state['storage']['data'])) {
      // No releases returned, so do search instead
      $results = discogs_search_for($term, $type, $api_key);
      if ($results) {
        $form_state['storage']['data'] = $results;
        $form_state['storage']['step'] = ($type == 'release') ? 2 : 1;
        $form_state['storage']['type'] = $type;
      }
      else {
        // Search returned nothing at all
        drupal_set_message(t('No releases found.'));
        $form_state['storage']['step'] = 0;
      }
    }
  }

  // Handle Step 1 (Select releases to import)
  elseif ($form_state['storage']['step'] == 1) {
    $releases = array();
    $data = empty($form_state['storage']['data']) ? array() : $form_state['storage']['data'];
    $type = $form_state['storage']['type'];
    foreach ($data as $did => $entry) {
      if (empty($entry)) continue;
      // Get data from Discogs.com
      $term = $did;
      $fetched = discogs_fetch_release($term, $type, $api_key);
      if ($fetched) {
        $releases = array_merge($releases, $fetched);
      }
    }
    $form_state['storage']['data'] = $releases;
    $form_state['storage']['step'] = 2;
  }

  // Handle Step 2 (Turn releases into nodes)
  elseif ($form_state['storage']['step'] == 2) {
    module_load_include('inc', 'node', 'node.pages');
    // Get submitted ID's
    $releases = $form_state['values']['data'];
    $releases = empty($releases) ? array() : array_values($releases);
    foreach ($releases as $release) {
      if (!empty($release) && ($release > 0)) {
        $array_data = discogs_fetch_release($release, 'release', $api_key);
        // Create form state from Discogs data
        $form_state                 = array();
        $form_state['values']       = discogs_parse_release($array_data);
        $form_state['track_count']  = count($form_state['values']['tracks']);
        // Create a new node
        $node = new stdClass();
        $node->type = 'discogs';
        // For some reason, I have to insert files into the node directly
        $node->files = _discogs_get_filenodes($form_state);
        drupal_execute('discogs_node_form', $form_state, $node);
      }
    }
    // Redirect to content admin page
    unset($form_state['storage']);
    $form_state['redirect'] = 'node/add/discogs_import';
  }
}

/**
 * Searches Discogs and returns an array of releases.
 *
 * For returning something other than releases, use discogs_search_for().
 *
 * @param $term
 *     Search term: release ID, artist or label name
 * @param $type
 *     Search type: 'release' (default), 'artist', 'label'
 * @return
 *     Multi-dimensional array containing all releases
 * @see discogs_search_for()
 */
function discogs_fetch_release($term, $type = 'release', $api_key = '') {
  if (empty($api_key)) {
    $error = t('You need to enter a Discogs API key into the <a href="@url">Discogs settings</a> page.',
    array('@url' => url('admin/settings/discogs')));
    drupal_set_message($error, 'error');
    return FALSE;
  }
  $term = (strpos($term, ' ') === FALSE) ? $term : urlencode($term);
  $url = "http://www.discogs.com/$type/$term?f=xml&api_key=$api_key";
  $xml_object = drupal_http_request($url, array('Accept-Encoding' => 'gzip'));
  return ($xml_object->code == '200') ? _discogs_xml_to_array($xml_object->data) : FALSE;
}

/**
 * Searches Discogs for a term and a type, returns an array of entries for that
 * type. The key will be the string needed to append to a Discogs REST URL.
 * This may be a URL-encoded string (for artists or labels), or a Discogs ID
 * (for releases).
 *
 * @param $term
 *     Discogs search term
 * @param $type
 *     Discogs search type (default 'release')
 * @return
 *     Array of releases matching the search terms
 * @see discogs_fetch_release()
 */
function discogs_search_for($term, $type = 'release', $api_key = '') {
  if (empty($api_key)) {
    $error = t('You need to enter a Discogs API key into the <a href="@url">Discogs settings</a>.',
      array('@url' => url('admin/settings/discogs')));
    drupal_set_message($error, 'error');
    return FALSE;
  }
  $xml_array = $rel_array = array();
  $searchtype = ($type == 'release') ? 'releases' : $type;
  $term = (strpos($term, ' ') === FALSE) ? $term : urlencode($term);
  $url = "http://www.discogs.com/search?type=$searchtype&q=$term&f=xml&api_key=$api_key";
  $xml_object = drupal_http_request($url, array('Accept-Encoding' => 'gzip'));
  $xml_string = ($xml_object->code == '200') ? $xml_object->data : FALSE;
  if ($xml_string) {
    $xml_array = _discogs_xml_to_array($xml_string);
    $xml_array = (is_array($xml_array['searchresults']['result'][0])
      ? $xml_array['searchresults']['result']
      : array($xml_array['searchresults']['result']));
  }
  else {
    $error = t('Error parsing Discogs XML data from %url.',
    array('%url' => $url));
    drupal_set_message($error, 'error');
    return FALSE;
  }
  // Turn search results into releases
  foreach ($xml_array as $rel) {
    if ($rel['type'] == 'release') {
      $this_rel['id'] = substr($rel['uri'], strrpos($rel['uri'], '/') + 1);
      $this_rel['title'] = $rel['title'];
      $this_rel['format'] = '';
      $this_rel['year'] = '';
      $rel_array[] = $this_rel;
    }
  }
  return $rel_array;
}

/**
 * Parses a single release into an array that is suitable to be passed to the
 * discogs_form() form handler as $form_state. We will also need to do
 * validation here, as it will not be done when the node is created.
 *
 * @param $releases
 *      Array of releases, passed from discogs_get()
 *      (only $releases[0] will be parsed!)
 * @param $cleanup
 *      Whether or not to clean up artist fields (default TRUE)
 * @return
 *      Multi-dimensional array to pass to discogs_form()
 * @see discogs_get()
 * @see _discogs_cleanup_artist_field()
 * @see discogs_form()
 */
function discogs_parse_release($releases, $cleanup = TRUE) {

  $release = $releases['release'];
  $new_form = array();

  // Theme artist field (in case of split releases)
  $discog_artist = theme('discogs_artists', $release['artists']['artist'], $cleanup);
  // Theme format array
  $discog_format = theme('discogs_formats', $release['formats']['format']);
  // Theme release's 'extraartists' as 'discog_credits'
  $discog_credits = theme('discogs_extra_artists', $release['extraartists'], $cleanup);

  // Get album info from $release array
  $new_form['title']          = $release['title'];
  $new_form['did']            = $release['id'];
  $new_form['discog_artist']  = strip_tags($discog_artist);
  $new_form['discog_label']   = $release['labels']['label']['name'];
  $new_form['discog_cat_num'] = $release['labels']['label']['catno'];
  $new_form['discog_format']  = strip_tags($discog_format);
  $new_form['discog_date']    = $release['released'];
  $new_form['discog_notes']   = $release['notes'];
  $new_form['discog_credits'] = $discog_credits;

  // Get array of tracks
  $tracks = $release['tracklist']['track'];
  // Take care of releases with only one track
  if (isset($tracks['title'])) {
    $tracks = array($tracks);
  }
  foreach ($tracks as $track) {
    $track_position = $track['position'];
    // Check if $position is empty, because sometimes Discogs "tracks"
    // aren't really tracks
    if (empty($track_position)) {
      unset($track);
      continue;
    }
    $track_artist = is_array($track['artists']) ? theme('discogs_artists', $track['artists']['artist'], $cleanup) : '' ;
    // Need to make sure title isn't more that 128 characters long
    $track_title = _discogs_trim($track['title']);
    // Theme track's 'extraartists' as 'track_notes'
    $track_notes = theme('discogs_extra_artists', $track['extraartists'], $cleanup);
    // For some reason, tracks without a duration return an empty array
    $track_duration = is_array($track['duration']) ? '' : $track['duration'];
    $tracks_array[] = array(
      'track_position' => $track_position,
      'track_artist'   => $track_artist,
      'track_title'    => $track_title,
      'track_duration' => $track_duration,
      'track_notes'    => $track_notes,
      'track_lyrics'   => ''
      );
  }
  $new_form['tracks'] = $tracks_array;

  // Handle release image
  $img_fields = _discogs_handle_images($release);
  $new_form = $new_form + $img_fields;

  $new_form['op'] = t('Save');
  return $new_form;
}

/**
 * Handles the images associated with a release.
 *
 * @param $release
 *      Array containing all information for a single release
 * @param $title
 *      Title of image node (used by Image module)
 * @return
 *      Array to insert into $form_state['values']
 * @see _discogs_copy_remote_file()
 * @see _discogs_copy_to_files()
 * @see _discogs_copy_to_images()
 */
function _discogs_handle_images($release) {
  $form    = array();
  $has_img = isset($release['images']);
  if ($has_img) {
    $urls = isset($release['images']['image']['uri']) ? array($release['images']['image']) : $release['images']['image'];
  }

  // Pass images to CCK imagefield module, if desired
  if (module_exists('imagefield') && (variable_get('discogs_img_handler', 0) == 2) && $has_img) {
    $form['field_discogs_cover'] = _discogs_copy_to_files($urls);
  }

  // Pass images to image_attach, if desired
  elseif (module_exists('image_attach') && (variable_get('discogs_img_handler', 0) == 3)) {
    $form['iids'] = $has_img ? _discogs_copy_to_images($urls, $release['title']) : array();
  }

  // Pass images to core upload module, if desired
  elseif (module_exists('upload') && (variable_get('discogs_img_handler', 0) == 1) && $has_img) {
    $copied_files = _discogs_copy_to_files($urls);
    // The files field expects an array with fid as key
    foreach ($copied_files as $file) {
      $files[$file['fid']] = $file + array(
        'description' => $file['filename'],
        'remove' => FALSE,
        'list' => 1,
        'new' => TRUE,
        'weight' => 0,
      );
    }
    $form['files'] = $files;
  }

  // If nothing else is available, just store the URL's in the node
  elseif ($has_img) {
    $discog_img            = _discogs_copy_remote_file($urls[0]['uri']);
    $discog_img150         = _discogs_copy_remote_file($urls[0]['uri150']);
    $form['discog_img']    = $discog_img ? $discog_img : '' ;
    $form['discog_img150'] = $discog_img150 ? $discog_img150 : '' ;
  }

  return $form;
}

/**
 * Takes an XML string, and returns a multidimensional array. Each "dimension"
 * may be a text string, a key-value array of tag attributes, and/or another
 * multidimensional array containing child nodes.
 *
 * @param $string
 *     String containing XML data to parse
 * @return
 *     Multidimensional array with attributes, values, and children
 */
function _discogs_xml_to_array(&$string) {
  $parser = xml_parser_create();
  xml_parser_set_option($parser, XML_OPTION_CASE_FOLDING, 0);
  xml_parse_into_struct($parser, $string, $vals, $index);
  xml_parser_free($parser);

  $multi_array = array();
  foreach ($vals as $val) {
    $tag = $val['tag'];
    if ($val['type'] == 'open') {
      if (isset($multi_array[$tag])) {
        if (isset($multi_array[$tag][0])) {
          $multi_array[$tag][]=array();
        }
        else {
          $multi_array[$tag]=array($multi_array[$tag], array());
        }
        $cv = &$multi_array[$tag][count($multi_array[$tag])-1];
      }
      else {
        $cv = &$multi_array[$tag];
      }
      if (isset($val['attributes'])) {
        foreach ($val['attributes'] as $k => $v) {
          $cv['_attributes'][$k] = $v;
        }
      }
      $cv['_nodes'] = array();
      $cv['_nodes']['_recursion'] = &$multi_array;
      $multi_array = &$cv['_nodes'];
    }
    elseif ($val['type'] == 'complete') {
      if (isset($multi_array[$tag])) {
        if (isset($multi_array[$tag][0])) {
          $multi_array[$tag][] = array();
        }
        else {
          $multi_array[$tag] = array($multi_array[$tag], array());
        }
        $cv = &$multi_array[$tag][count($multi_array[$tag])-1];
      }
      else {
        $cv = &$multi_array[$tag];
      }
      if (isset($val['attributes'])) {
        foreach ($val['attributes'] as $k => $v) {
          $cv['_attributes'][$k] = $v;
        }
      }
      if (isset($val['value'])) {
        $cv['value'] = trim($val['value']);
      }
    }
    elseif ($val['type'] == 'close') {
      $multi_array = &$multi_array['_recursion'];
    }
  }
  // Clean up the recursion and merge attributes with child nodes
  $multi_array = _discogs_xml_clean($multi_array);

  // Make sure everything's OK
  $all_ok = $multi_array['resp']['stat'];
  if ($all_ok != 'ok') {
    $error = t('Discogs.com response error: %err', array('%err' => $all_ok));
    drupal_set_message($error, 'error');
    return FALSE;
  }
  return $multi_array['resp'];
}

/**
 * Helper function for _discogs_xml_to_array.
 *
 * Cleans up the array structure, by getting rid of the recursive data, and
 * merging the values, attributes and child nodes.
 *
 * @param array $old_array
 *   Old array to clean up
 * @return array $new_array
 *   Cleaned up array
 */
function _discogs_xml_clean($old_array) {
  // Double check that it's an array, for recursion
  if (!is_array($old_array)) {
    return $old_array;
  }
  foreach ($old_array as $k => $v) {
    if ($k === '_recursion') {
      unset($old_array[$k]);
    }
    elseif (is_array($old_array[$k]['_nodes']) && is_array($old_array[$k]['_attributes'])) {
      $old_nodes = _discogs_xml_clean($old_array[$k]['_nodes']);
      $new_merged = array_merge($old_array[$k]['_attributes'], $old_nodes);
      $new_array[$k] = $new_merged;
    }
    elseif (isset($old_array[$k]['value']) && is_array($old_array[$k]['_attributes'])) {
      $value['value'] = trim($old_array[$k]['value']);
      $new_merged = array_merge($old_array[$k]['_attributes'], $value);
      $new_array[$k] = $new_merged;
    }
    elseif (is_array($old_array[$k]['_nodes'])) {
      $new_array[$k] = _discogs_xml_clean($old_array[$k]['_nodes']);
    }
    elseif (isset($old_array[$k]['_attributes'])) {
      $new_array[$k] = $old_array[$k]['_attributes'];
    }
    elseif (isset($old_array[$k]['value'])) {
      $new_array[$k] = $old_array[$k]['value'];
    }
    else {
      $old_nodes = _discogs_xml_clean($old_array[$k]);
      $new_array[$k] = $old_nodes;
    }
  }
  return $new_array;
}

/**
 * Copies a binary file from a remote URL to the sub-directory of
 * file_directory_path() set by the discogs_img_dir variable.
 *
 * @param $url
 *      URL of file to copy
 * @return
 *      URL of current filename (if copying is successful), or FALSE (if not)
 */
function _discogs_copy_remote_file($url) {
  if (empty($url)) {
    return FALSE;
  }
  @$file = fopen($url, "rb");
  if (!$file) {
    $error = t("Failed to open %url", array('%url' => $url));
    drupal_set_message($error, 'error');
    return FALSE;
  }
  else {
    $discogs_img_dir = variable_get('discogs_img_dir', '');
    $files_subdir = empty($discogs_img_dir) ? '' : rtrim($discogs_img_dir, '/\\') . '/';
    $to = file_directory_path() . '/' . $files_subdir . basename($url);
    @$fc = fopen($to, "wb");
    if (!$fc) {
      $error = t("Could not write to %to", array('%to' => $to));
      drupal_set_message($error, 'error');
      return FALSE;
    }
    while (!feof($file)) {
      $line = fread($file, 1028);
      fwrite($fc, $line);
    }
    fclose($fc);
    return $to;
  }
}

/**
 * Takes an array of URL's, copies them to the files directory, and returns
 * an array suitable for inserting into $form_state.
 *
 * @param $urls
 *   Array of URL's to import
 * @return
 *   Array to insert into $form_state
 */
function _discogs_copy_to_files($urls) {
  $i = 0;
  $files = array();
  global $user;
  foreach ($urls as $url) {
    $img_url = _discogs_copy_remote_file($url['uri']);
    if ($img_url) {
      // Create file object and update files table
      $file            = new stdClass();
      $file->uid       = $user->uid;
      $file->filename  = basename($img_url);
      $file->filepath  = $img_url;
      $file->filemime  = mime_content_type(basename($img_url));
      $file->filesize  = filesize($img_url);
      $file->status    = 0;
      $file->timestamp = time();
      drupal_write_record('files', $file);

      // Create array for form submission
      $files[$i]            = (array) $file;
      $files[$i]['_weight'] = $i;
      $i++;
    }
  }
  return $files;
}

/**
 * Helper function to create image nodes, and return an array of their nid's
 * to put into $form_state['values']['iids']
 *
 * @param $urls
 *   Array of (local) URL's to images
 * @param $title
 *   Title of the release
 * @return
 *   Array of image node nid's
 */
function _discogs_copy_to_images($urls, $title = '') {
  // Need an iterator to make sure we don't get too many images
  $i = 1;
  $max = variable_get('image_attach_maximum_discogs', 0);
  foreach ($urls as $url) {
    if (($i > $max) && ($max > 0)) {
      continue;
    }
    $img_url = _discogs_copy_remote_file($url['uri']);
    if ($img_url) {
      $new_title = empty($title) ? t('Release image @i', array('@i' => $i)) : t('@title image @i', array('@title' => $title, '@i' => $i));
      $node = image_create_node_from($img_url, $new_title);
      if ($node) {
        $form[] = $node->nid;
      }
      else {
        $error = t("Could not create image node from %url",
        array('%url' => $img_url));
        drupal_set_message($error, 'error');
      }
    }
    $i++;
  }
  return $form;
}

/**
 * Takes $form_state, and returns an array of file node objects to be inserted
 * into $node->files.
 *
 * @param $form_state
 *   Drupal form state
 * @return
 *   Array of file node objects
 */
function _discogs_get_filenodes($form_state) {
  $filenodes = array();
  if (is_array($form_state['values']['files'])) {
    foreach ($form_state['values']['files'] as $file) {
      $filenodes[] = (object) $file;
    }
  }
  return $filenodes;
}